==> hlog-0.10-rc1.tar.gz: hlog-0.10-rc1/country.c <==
/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/types.h>
#include <jeffpc/io.h>
#include <jeffpc/rbtree.h>
#include <jeffpc/val.h>
#include <jeffpc/cstr.h>
#include <jeffpc/mem.h>

#include "country.h"
#include "util.h"

struct country {
	struct str *prefix;	/* e.g., "K" */
	struct str *name;	/* e.g., "United States" */
	struct str *continent;	/* e.g., "NA" */
	double lat;		/* e.g., 37.53 (degrees, N is positive) */
	double lon;		/* e.g., -91.67 (degrees, E is positive) */
	int32_t tz;		/* e.g., -18000 (seconds, E is positive) */
	uint16_t waz;		/* e.g., 5 */
	uint16_t itu;		/* e.g., 8 */

	struct country *alias;	/* what is this an alias for */

	struct rb_node node;
};

struct rb_tree countries;

static int country_cmp(const void *va, const void *vb)
{
	const struct country *a = va;
	const struct country *b = vb;

	return strcmp(str_cstr(a->prefix), str_cstr(b->prefix));
}

/*
 * This comment is based on text from:
 * http://www.country-files.com/cty-dat-format/
 *
 * cty.csv format:
 *   - Primary DXCC Prefix (A “*” preceding this prefix indicates that the
 *     country is on the DARC WAEDC list, and counts in CQ-sponsored
 *     contests, but not ARRL-sponsored contests)
 *   - Country Name
 *   - ??? (291 for 'K')
 *   - 2-letter continent abbreviation
 *   - CQ Zone
 *   - ITU Zone
 *   - Latitude in degrees, + for North
 *   - Longitude in degrees, + for West
 *   - Local time offset from GMT
 *   - Alias DXCC prefixes (including the primary one), separated by commas.
 *     A semi-colon (;) terminates the last alias prefix in the list.
 *
 *   If an alias prefix begins with a '=', this indicates that the prefix is
 *   to be treated as a full callsign, i.e. must be an exact match.
 *
 *   The following special characters can be applied after an alias prefix:
 *
 *   (#)	Override CQ Zone
 *   [#]	Override ITU Zone
 *   <#/#>	Override latitude/longitude
 *   {aa}	Override Continent
 *   ~#~	Override local time offset from GMT
 */

static int add_aliases(char *aliases, struct country *parent)
{
	struct country *alias;
	struct country *existing;
	char *prefix;

	while ((prefix = strsep(&aliases, " "))) {
		char *modifiers;
		int ret;

		/* the list of aliases includes the primary as well */
		if (!strcmp(prefix, str_cstr(parent->prefix)))
			continue;

		/* skip over exact callsigns for now */
		if (*prefix == '=')
			continue;

		alias = malloc(sizeof(struct country));
		if (!alias)
			return -ENOMEM;

		modifiers = strpbrk(prefix, "([<{~");

		/* default to exact copy except for the prefix */
		alias->prefix = STR_DUP_LEN(prefix, modifiers - prefix);
		alias->name = str_getref(parent->name);
		alias->continent = str_getref(parent->continent);
		alias->lat = parent->lat;
		alias->lon = parent->lon;
		alias->tz = parent->tz;
		alias->waz = parent->waz;
		alias->itu = parent->itu;
		alias->alias = parent;

		ret = 0;

		/* parse the modifiers */
		while (modifiers && *modifiers && !ret) {
			const char *endsep;
			char *this;
			void *val;

			switch (*modifiers) {
				case '(':
					endsep = ")";
					val = &alias->waz;
					break;
				case '[':
					endsep = "]";
					val = &alias->itu;
					break;
				case '<':
					endsep = ">";
					val = NULL;
					break;
				case '{':
					endsep = "}";
					val = NULL;
					break;
				case '~':
					endsep = "~";
					val = NULL;
					break;
			}

			this = strsep(&modifiers, endsep);

			if ((*this == '(') || (*this == '[')) {
				ret = str2u16(this + 1, val);
			} else if (*this == '{') {
				str_putref(alias->continent);
				alias->continent = STR_DUP(this + 1);
			} else if (*this == '<') {
				fprintf(stderr, "country: Ignoring %s alias "
					"lat/lon override %s for %s\n",
					str_cstr(alias->prefix), this + 1,
					str_cstr(parent->prefix));
			} else if (*this == '~') {
				fprintf(stderr, "country: Ignoring %s alias "
					"timezone override %s for %s\n",
					str_cstr(alias->prefix), this + 1,
					str_cstr(parent->prefix));
			}
		}

		if (ret) {
			fprintf(stderr, "country: Failed to parse %s alias for "
				"%s - bad modifiers, ignoring alias\n",
				str_cstr(alias->prefix),
				str_cstr(parent->prefix));

			str_putref(alias->prefix);
			str_putref(alias->name);
			str_putref(alias->continent);
			free(alias);
			continue;
		}

		existing = rb_insert(&countries, alias);
		if (existing) {
			bool ealias = existing->alias != NULL;

			fprintf(stderr, "country: Tried to add %s as alias for "
				"%s, but %s already exists%s%s - ignoring\n",
				str_cstr(alias->prefix),
				str_cstr(parent->prefix),
				str_cstr(alias->prefix),
				ealias ? " as an alias for " : "",
				ealias ? str_cstr(existing->alias->prefix) : "");
		}
	}

	return 0;
}

static int parse_country(char **raw)
{
	struct country *country;
	char *field;
	int ret;

	country = zalloc(sizeof(struct country));
	if (!country)
		return -ENOMEM;

	/* Primary DXCC prefix */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	country->prefix = STR_DUP(field);

	/* Country name */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	country->name = STR_DUP(field);

	/* Mystery int */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	/* ignore */

	/* Continent */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	country->continent = STR_DUP(field);

	/* CQ zone */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	ret = str2u16(field, &country->waz);
	if (ret)
		goto err_free;

	/* ITU zone */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	ret = str2u16(field, &country->itu);
	if (ret)
		goto err_free;

	/* Latitude */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	ret = extract_float_as_double(field, &country->lat);
	if (ret)
		goto err_free;

	/* Longitude */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	ret = extract_float_as_double(field, &country->lon);
	if (ret)
		goto err_free;

	country->lon *= -1; /* E should be positive */

	/* Timezone */
	field = strsep(raw, ",");
	if (!field)
		goto err_free;

	ret = extract_float_as_int32(field, 3600, &country->tz);
	if (ret)
		goto err_free;

	country->tz *= -1; /* E should be positive */

	/* NOTE: we are adding this before we add any of the aliases */
	rb_add(&countries, country);

	/* Process aliases */
	field = strsep(raw, "\n");

	/* trim EOL related chars */
	if (field[strlen(field) - 1] == '\r')
		field[strlen(field) - 1] = '\0';

	if (field[strlen(field) - 1] != ';')
		goto err;

	field[strlen(field) - 1] = '\0';

	ret = add_aliases(field, country);
	if (ret)
		goto err;

	return 0;

err_free:
	str_putref(country->prefix);
	str_putref(country->name);
	str_putref(country->continent);
	free(country);

err:
	return -EINVAL;
}

int country_load(const char *fname)
{
	char *raw;
	int ret;

	raw = read_file(fname);
	if (IS_ERR(raw))
		return PTR_ERR(raw);

	rb_create(&countries, country_cmp, sizeof(struct country),
		  offsetof(struct country, node));

	for (char *tmp = raw; tmp && *tmp; ) {
		ret = parse_country(&tmp);
		if (ret)
			break;
	}

	free(raw);

	return ret;
}

bool country_lookup(struct str *callsign, struct country_result *result)
{
	char tmp[32];
	struct str tmpstr = STR_STATIC_INITIALIZER(tmp);
	struct country key = {
		.prefix = &tmpstr,
	};
	size_t len;

	if (!callsign)
		return false;

	strcpy_safe(tmp, str_cstr(callsign), sizeof(tmp));
	len = strlen(tmp);

	while (len) {
		struct country *country;

		country = rb_find(&countries, &key, NULL);
		if (country) {
			result->prefix = str_getref(country->prefix);
			result->name = str_getref(country->name);
			result->continent = str_getref(country->continent);
			result->lat = country->lat;
			result->lon = country->lon;
			result->tz = country->tz;
			result->waz = country->waz;
			result->itu = country->itu;

			return true;
		}

		tmp[--len] = '\0';
	}

	// https://en.wikipedia.org/wiki/Amateur_radio_call_signs#A_unique_international_prefix

	return false;
}

==> hlog-0.11-rc1.tar.gz: hlog-0.11-rc1/country.c <==
/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/types.h>
#include <jeffpc/io.h>
#include <jeffpc/rbtree.h>
#include <jeffpc/val.h>
#include <jeffpc/cstr.h>
#include <jeffpc/mem.h>

#include "country.h"
#include "util.h"

struct country {
	struct str *prefix;	/* e.g., "AC" */
	struct str *name;	/* e.g., "K" */
	struct str *fullname;	/* e.g., "United States" */
	struct str *continent;	/* e.g., "NA" */
	double lat;		/* e.g., 37.53 (degrees, N is positive) */
	double lon;		/* e.g., -91.67 (degrees, E is positive) */
	int32_t tz;		/* e.g., -18000 (seconds, E is positive) */
	uint16_t waz;		/* e.g., 5 */
	uint16_t itu;		/* e.g., 8 */
	uint16_t code;		/* e.g., 291 */

	struct rb_node node;
};

struct rb_tree countries;

static int country_cmp(const void *va, const void *vb)
{
	const struct country *a = va;
	const struct country *b = vb;

	return strcmp(str_cstr(a->prefix), str_cstr(b->prefix));
}

/*
 * This comment is based on text from:
 * http://www.country-files.com/cty-dat-format/
 *
 * cty.csv format:
 *   - "Printable" DXCC Prefix (A “*” preceding this prefix indicates that
 *     the country is on the DARC WAEDC list, and counts in CQ-sponsored
 *     contests, but not ARRL-sponsored contests)
 *   - Country Name
 *   - Country Code (e.g., 291 for 'K')
 *   - 2-letter continent abbreviation
 *   - CQ Zone
 *   - ITU Zone
 *   - Latitude in degrees, + for North
 *   - Longitude in degrees, + for West
 *   - Local time offset from GMT
 *   - Prefixes, separated by commas.  A semi-colon (;) terminates the list.
 *
 *   If a prefix begins with a '=', this indicates that the prefix is to be
 *   treated as a full callsign, i.e. must be an exact match.
 *
 *   The following special characters can be applied after a prefix:
 *
 *   (#)	Override CQ Zone
 *   [#]	Override ITU Zone
 *   <#/#>	Override latitude/longitude
 *   {aa}	Override Continent
 *   ~#~	Override local time offset from GMT
 */

static int add_prefixes(char *prefixes, struct country *template)
{
	struct country *country;
	struct country *existing;
	char *prefix;

	while ((prefix = strsep(&prefixes, " "))) {
		char *modifiers;
		int ret;

		/* skip over exact callsigns for now */
		if (*prefix == '=')
			continue;

		country = malloc(sizeof(struct country));
		if (!country)
			return -ENOMEM;

		modifiers = strpbrk(prefix, "([<{~");

		/* default to exact copy except for the prefix */
		country->prefix = STR_DUP_LEN(prefix, modifiers - prefix);
		country->name = str_getref(template->name);
		country->fullname = str_getref(template->fullname);
		country->continent = str_getref(template->continent);
		country->lat = template->lat;
		country->lon = template->lon;
		country->tz = template->tz;
		country->waz = template->waz;
		country->itu = template->itu;
		country->code = template->code;

		ret = 0;

		/* parse the modifiers */
		while (modifiers && *modifiers && !ret) {
			const char *endsep;
			char *this;
			void *val;

			switch (*modifiers) {
				case '(':
					endsep = ")";
					val = &country->waz;
					break;
				case '[':
					endsep = "]";
					val = &country->itu;
					break;
				case '<':
					endsep = ">";
					val = NULL;
					break;
				case '{':
					endsep = "}";
					val = NULL;
					break;
				case '~':
					endsep = "~";
					val = NULL;
					break;
			}

			this = strsep(&modifiers, endsep);

			if ((*this == '(') || (*this == '[')) {
				ret = str2u16(this + 1, val);
			} else if (*this == '{') {
				str_putref(country->continent);
				country->continent = STR_DUP(this + 1);
			} else if (*this == '<') {
				fprintf(stderr, "country: Ignoring %s prefix "
					"lat/lon override %s for %s\n",
					str_cstr(country->prefix), this + 1,
					str_cstr(template->name));
			} else if (*this == '~') {
				fprintf(stderr, "country: Ignoring %s prefix "
					"timezone override %s for %s\n",
					str_cstr(country->prefix), this + 1,
					str_cstr(template->name));
			}
		}

		if (ret) {
			fprintf(stderr, "country: Failed to parse %s prefix "
				"for %s - bad modifiers, ignoring\n",
				str_cstr(country->prefix),
				str_cstr(template->name));

			str_putref(country->prefix);
			str_putref(country->name);
			str_putref(country->fullname);
			str_putref(country->continent);
			free(country);
			continue;
		}

		existing = rb_insert(&countries, country);
		if (existing)
			fprintf(stderr, "country: Tried to add %s prefix for "
				"%s, but %s already exists for %s - ignoring\n",
				str_cstr(country->prefix),
				str_cstr(template->name),
				str_cstr(country->prefix),
				str_cstr(existing->name));
	}

	return 0;
}

static int parse_country(char **raw)
{
	struct country template;
	char *field;
	int ret;

	memset(&template, 0, sizeof(struct country));

	/* Primary DXCC prefix */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	template.name = STR_DUP(field);

	/* Country name */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	template.fullname = STR_DUP(field);

	/* Country code */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	ret = str2u16(field, &template.code);
	if (ret)
		goto err;

	/* Continent */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	template.continent = STR_DUP(field);

	/* CQ zone */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	ret = str2u16(field, &template.waz);
	if (ret)
		goto err;

	/* ITU zone */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	ret = str2u16(field, &template.itu);
	if (ret)
		goto err;

	/* Latitude */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	ret = extract_float_as_double(field, &template.lat);
	if (ret)
		goto err;

	/* Longitude */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	ret = extract_float_as_double(field, &template.lon);
	if (ret)
		goto err;

	template.lon *= -1; /* E should be positive */

	/* Timezone */
	field = strsep(raw, ",");
	if (!field)
		goto err;

	ret = extract_float_as_int32(field, 3600, &template.tz);
	if (ret)
		goto err;

	template.tz *= -1; /* E should be positive */

	/* Process prefixes */
	field = strsep(raw, "\n");

	/* trim EOL related chars */
	if (field[strlen(field) - 1] == '\r')
		field[strlen(field) - 1] = '\0';

	if (field[strlen(field) - 1] != ';')
		goto err;

	field[strlen(field) - 1] = '\0';

	ret = add_prefixes(field, &template);
	if (ret)
		goto err;

	ASSERT3P(template.prefix, ==, NULL);
	str_putref(template.name);
	str_putref(template.fullname);
	str_putref(template.continent);

	return 0;

err:
	ASSERT3P(template.prefix, ==, NULL);
	str_putref(template.name);
	str_putref(template.fullname);
	str_putref(template.continent);

	return -EINVAL;
}

int country_load(int dirfd, const char *fname)
{
	char *raw;
	int ret;

	raw = read_file_at(dirfd, fname);
	if (IS_ERR(raw))
		return PTR_ERR(raw);

	rb_create(&countries, country_cmp, sizeof(struct country),
		  offsetof(struct country, node));

	for (char *tmp = raw; tmp && *tmp; ) {
		ret = parse_country(&tmp);
		if (ret)
			break;
	}

	free(raw);

	return ret;
}

bool country_lookup(struct str *callsign, struct country_result *result)
{
	char tmp[32];
	struct str tmpstr = STR_STATIC_INITIALIZER(tmp);
	struct country key = {
		.prefix = &tmpstr,
	};
	size_t len;

	if (!callsign)
		return false;

	strcpy_safe(tmp, str_cstr(callsign), sizeof(tmp));
	len = strlen(tmp);

	while (len) {
		struct country *country;

		country = rb_find(&countries, &key, NULL);
		if (country) {
			result->prefix = str_getref(country->prefix);
			result->name = str_getref(country->name);
			result->fullname = str_getref(country->fullname);
			result->continent = str_getref(country->continent);
			result->lat = country->lat;
			result->lon = country->lon;
			result->tz = country->tz;
			result->waz = country->waz;
			result->itu = country->itu;
			result->code = country->code;

			return true;
		}

		tmp[--len] = '\0';
	}

	// https://en.wikipedia.org/wiki/Amateur_radio_call_signs#A_unique_international_prefix

	return false;
}
